import setuptools

setuptools.setup(
    name="schematacode",
    version="1.0.0",
    author="B. T. Milnes",
    description="Schemata is a file format that makes writing XML and JSON schemas easier.",
    long_description="",
    url="https://github.com/SchemataCode/Schemata",
    packages=["schemata"]
)
